#--
#
#################################################################################
# Copyright 2011 CENATIC y IEPALA
#
# Licencia con arreglo a la EUPL, Versión 1.1 o –en cuanto sean aprobadas por la Comisión Europea– 
# versiones posteriores de la EUPL (la «Licencia»);
# Solo podrá usarse esta obra si se respeta la Licencia.
# Puede obtenerse una copia de la Licencia en:
#
# http://www.osor.eu/eupl/european-union-public-licence-eupl-v.1.1
#
# Salvo cuando lo exija la legislación aplicable o se acuerde por escrito, 
# el programa distribuido con arreglo a la Licencia se distribuye «TAL CUAL»,
# SIN GARANTÍAS NI CONDICIONES DE NINGÚN TIPO, ni expresas ni implícitas.
# Véase la Licencia en el idioma concreto que rige los permisos y limitaciones que establece la Licencia.
#################################################################################
#
#++
class Presupuesto < ActiveRecord::Base
  untranslate_all
  belongs_to :moneda
  belongs_to :partida
  belongs_to :libro
  belongs_to :agente
  belongs_to :subpartida
  belongs_to :proyecto
  belongs_to :etapa
  belongs_to :marcado

  has_many :presupuesto_x_agente, :dependent => :destroy
  has_many :financiador, :source => :agente, :through => :presupuesto_x_agente
  has_many :presupuesto_x_actividad, :dependent => :destroy
  has_many :actividad, :through => :presupuesto_x_actividad

  has_many :presupuesto_x_actividad, :dependent => :destroy
  has_many :presupuesto_detallado, :dependent => :destroy
  has_many :comentario, :dependent => :destroy

  validates_presence_of :moneda, :message => _("Moneda") + _(" no puede estar vacío.")
  #validates_presence_of :proyecto, :message => _("Proyecto") + _(" no puede estar vacío.")
  validates_presence_of :agente, :message => _("Agente") + _(" no puede estar vacío.")
  validates_presence_of :partida, :message => _("Partida") + _(" no puede estar vacío.")
  validates_presence_of :concepto, :message => _("Concepto") + _(" no puede estar vacío.")
  validates_presence_of :etapa, :message => _("Etapa") + _(" no puede estar vacío.")
  validates_numericality_of :importe, :greater_than => 0, :message => _("Importe") + _(" no puede estar vacío.")
  validates_numericality_of :numero_unidades, :greater_than => 0, :message => _("Numero unidades") + _(" no puede estar vacío.")

  validates_uniqueness_of :concepto, :scope => [:proyecto_id], :message => _("Concepto ") + "'%{value}'" + _(" repetido")

  validate :relacion_partida_subpartida
  before_save :adaptacion_datos
  before_destroy :verificar_subpartida
  after_create :dividir_por_mes


  def verificar_subpartida
    ( subpartida.destroy if (subpartida.presupuesto -  [ self ]).empty? ) if subpartida
  end

  def adaptacion_datos
    # Concepto en mayusculas
    self.concepto = self.concepto.upcase
    # Cambia la tasa de cambio si es nuevo o se ha actualizado etapa o moneda
    if (self.id.nil? || self.etapa_id_changed? || self.moneda_id_changed?)
      objeto = (self.proyecto || self.agente)
      tc = TasaCambio.tasa_cambio_para_presupuesto(self.etapa,self.moneda_id)
      self.tasa_cambio_id = tc ? tc.id : nil
    end
  end
  
  def subpartida_nombre
    subpartida.nombre if subpartida
  end 

  # Comprueba que la subpartida no este asociada a otra partida distinta de la actual
  def relacion_partida_subpartida
    if proyecto_id
      (Presupuesto.find(:all, :conditions => [ "proyecto_id = ? AND partida_id != ?", self.proyecto.id, self.partida_id ]) - [self]).each do |presupuesto|
        errors.add("Subpartida", _("La subpartida está relacionada con otra partida distinta de la asignada")) if presupuesto.subpartida && presupuesto.subpartida.nombre == self.subpartida.nombre
      end if (self.subpartida and self.subpartida != "")
    elsif proyecto_id.nil?
      (Presupuesto.find(:all, :conditions => [ "agente_id = ? AND partida_id != ?", self.agente.id, self.partida_id ]) - [self]).each do |presupuesto|
        errors.add("Subpartida", _("La subpartida está relacionada con otra partida distinta de la asignada")) if presupuesto.subpartida && presupuesto.subpartida.nombre == self.subpartida.nombre
      end if (self.subpartida and self.subpartida != "")
    end
  end

  def subpartida_nombre=(nombre)
    if nombre.blank?
      self.subpartida_id = nil 
    else
      self.subpartida = Subpartida.find_by_nombre(nombre)  
      self.subpartida ||= Subpartida.new(:nombre => nombre)
    end
  end 

  #Metodo especifico creado para el ruport y la exportacion
  def moneda_abreviatura
    return Moneda.find(self.moneda_id).abreviatura if self.moneda_id
  end


  #Metodo especifico creado para el ruport y la exportacion
  def agente_nombre
    return Agente.find(self.agente_id).nombre if self.agente_id
  end

  # Devuelve el importe en la moneda base (principal) del proyecto/agente
  def importe_en_base
    objeto = self.proyecto || self.agente
    ixtc = self.importe * self.tasa_cambio
    return ('%.2f' % ixtc).to_s + " " + objeto.moneda_principal.abreviatura if ixtc
  end

  # Devuelve un array con los importes por actividad
  def importes_por_actividades
    salida = Array.new
    pxas = presupuesto_x_actividad.all
    pxas.each {|pxa| salida.push( pxa.actividad.codigo + ": " + pxa.importe.to_s ) }
    return salida
  end

  # Devuelve un array con los importes por financiador
  def importes_por_financiadores
    salida = Array.new
    pxfs = presupuesto_x_agente.all
    pxfs.each {|pxf| salida.push( pxf.agente.nombre + ": " + pxf.importe.to_s ) }
    return salida
  end



  # Actualiza el listado de agentes con los que se ha asociado el presupuesto para un determinado proyecto.
  def actualizar_presupuesto_x_agente listado
    presupuesto_x_agente.clear
    listado.each do |key, value|
      pxa = presupuesto_x_agente.create(value)
      errors.add("", pxa.errors.inject('') {|total, e| total + e[1] + " " }) unless pxa.errors.empty?
    end        
    if listado.inject(0) {|suma, value| suma + moneda_a_float(value.pop[:importe_convertido])} != self.importe
      errors.add(_('Importes'), _("Los importes de los financiadores no suman el importe total del presupuesto."))        
    end
  end

  # Actualiza el listado de actividades con los que se ha asociado el presupuesto para un determinado proyecto.
  def actualizar_presupuesto_x_actividad listado
    presupuesto_x_actividad.clear
    listado.each do |key, value|
      pxa = presupuesto_x_actividad.create(value)
      errors.add("", pxa.errors.inject('') {|total, e| total + e[1] + " " }) unless pxa.errors.empty?
    end
    if  listado.inject(0) {|suma, value| suma + moneda_a_float(value.pop[:importe_convertido])} != self.importe
      errors.add(_('Importes'), _("Los importes de las actividades no suman el importe total del presupuesto."))        
    end
  end

  # Actualiza el detalle del presupuesto.
  def actualizar_presupuesto_detallado listado
    presupuesto_detallado.clear
    listado.each do |mes, value|
      presupuesto_detallado.create(value) unless value[:importe] == ""
    end
    errors.add(_('Porcentajes'),_("La suma del detalle no coincide con la importe de la linea del presupuesto. Vuelva a editar el detalle para corregirlo.")) unless self.importe == presupuesto_detallado.inject(0) {|sum,p| sum+p.importe}
  end

  def dividir_por_actividades actividades
     detalle = Hash.new
     importe_x_actividad, resto_importe = importe.divmod(actividades.size)
     numero_unidades_x_actividad, resto_numero_unidades = numero_unidades.divmod(actividades.size)
     for actividad in actividades
        if actividad == actividades.last
          importe_x_actividad += resto_importe
          numero_unidades_x_actividad += resto_numero_unidades
        end
        detalle[actividad.id] = { :importe_convertido => importe_x_actividad, :numero_unidades => numero_unidades_x_actividad , :actividad_id => actividad.id }  
     end
     actualizar_presupuesto_x_actividad detalle
  end

  def dividir_por_mes fecha_inicio=nil, fecha_fin=nil
     fecha_inicio ||= self.etapa.fecha_inicio if self.etapa
     fecha_fin ||= self.etapa.fecha_fin if self.etapa
     detalle, fecha = Hash.new, fecha_inicio
     importe_x_mes, resto = (importe * 100).divmod(etapa.meses)
     importe_x_mes, resto = importe_x_mes/100, resto/100
     for mes in 0..(etapa.meses - 1)
        detalle[(mes + 1)] = { :importe => importe_x_mes + resto, :fecha_inicio => fecha >> mes, :fecha_fin => (fecha >> mes + 1) - 1, :mes => (mes + 1) }  if (mes + 1) == etapa.meses
        detalle[(mes + 1)] = { :importe => importe_x_mes, :fecha_inicio => fecha >> mes, :fecha_fin => (fecha >> mes + 1) - 1, :mes => (mes + 1) } unless (mes + 1)  == etapa.meses
     end
     actualizar_presupuesto_detallado detalle
  end

  # utilizado solamente en el partial comunes/_activ_financ.rhtml que ya no se usa
  def comprobar_financiadores
    return ( presupuesto_x_agente.inject(0) {|suma, f| suma + f.importe} == self.importe ) ? true : false
  end

  def comprobar_actividades
    return ( presupuesto_x_actividad.inject(0) {|suma, f| suma + f.importe} == self.importe ) ? true : false
  end

  def todas_actividades?
    return presupuesto_x_actividad.size == proyecto.actividad.size if proyecto
  end

  # Comprueba si el detalle del presupuesto se ha configurado adecuadamente.
  def comprobar_presupuesto_detallado
    return ( presupuesto_detallado.inject(0) {|suma, p| suma + p.importe} == self.importe ) ? true : false
  end

  # Devuelve la tasa de cambio para la línea de presupuesto (si existe)
  def tasa_cambio
    tc = TasaCambio.find_by_id(self.tasa_cambio_id)
    return (tc ? tc.tasa_cambio : 0.0)
  end

  #--
  # FINDERS complicados... Mas alla de ActiveRecordBase
  # ++

  # Busqueda creada manualmente para los detalles de busquedas especificas que no cubre el find habitual de ActiveRecord.
  def self.busqueda campos, condiciones, asociacion, agrupado, ordenado
    consulta = " SELECT " + campos + " FROM presupuesto  "
    consulta += asociaciones asociacion
    consulta += " WHERE " + merge_conditions(condiciones) if condiciones
    consulta += " GROUP BY " + agrupado if agrupado
    consulta += " ORDER BY " + ordenado if ordenado
    return find_by_sql consulta 
  end
  
  # Cutrada para sustituir el include para el metodo busqueda.
  def self.asociaciones asociacion
    asociacion.inject(' ') do |suma, elemento|
      case elemento
        # when :presupuesto_x_proyecto
        # suma + ' LEFT JOIN presupuesto_x_proyecto on presupuesto_x_proyecto.presupuesto_id = presupuesto.id'
        #when :tasa_cambio_presupuesto
        #suma + ' LEFT JOIN tasa_cambio_presupuesto on tasa_cambio_presupuesto.presupuesto_id = presupuesto.id'
        when :presupuesto_x_agente
        suma + ' LEFT JOIN presupuesto_x_agente on presupuesto_x_agente.presupuesto_id = presupuesto.id'
        when :presupuesto_x_actividad
        suma + ' LEFT JOIN presupuesto_x_actividad on presupuesto_x_actividad.presupuesto_id = presupuesto.id'
        when :partida_x_partida_financiacion
        suma + ' LEFT JOIN partida_x_partida_financiacion on partida_x_partida_financiacion.partida_id = presupuesto.partida_id'
        when :partida_financiacion
        suma + ' LEFT JOIN partida_financiacion on partida_financiacion.id = partida_x_partida_financiacion.partida_financiacion_id'
        when :presupuesto_detallado
        suma + ' LEFT JOIN presupuesto_detallado on presupuesto_detallado.presupuesto_id = presupuesto.id'
      end
    end
  end



  # Metodo que iguala algunos valores basicos para procesos de insercion de datos.
  def self.datos_basicos_igualar presupuesto, presupuesto_anterior
    presupuesto.partida = presupuesto_anterior.partida
    presupuesto.etapa_id = presupuesto_anterior.etapa_id
    presupuesto.agente_id = presupuesto_anterior.agente_id
    presupuesto.moneda_id = presupuesto_anterior.moneda_id
    presupuesto.importe = presupuesto_anterior.importe
    presupuesto.coste_unitario = presupuesto_anterior.coste_unitario
    presupuesto.numero_unidades = presupuesto_anterior.numero_unidades
    presupuesto.unidad = presupuesto_anterior.unidad
    presupuesto.subpartida_id = presupuesto_anterior.subpartida_id
  end

end
